"use strict";
/*
 * Copyright 2022 Fernando Boucquez
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.BootstrapUtils = exports.KnownError = void 0;
const fs_1 = require("fs");
const Handlebars = require("handlebars");
const _ = require("lodash");
const os_1 = require("os");
const path_1 = require("path");
const symbol_sdk_1 = require("symbol-sdk");
const CryptoUtils_1 = require("./CryptoUtils");
const Utils_1 = require("./Utils");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const yaml = require('js-yaml');
class KnownError extends Error {
    constructor() {
        super(...arguments);
        this.known = true;
    }
}
exports.KnownError = KnownError;
class BootstrapUtils {
    static toAns1(privateKey) {
        const prefix = '302e020100300506032b657004220420';
        return `${prefix}${privateKey.toLowerCase()}`;
    }
    static sleep(ms) {
        // Create a promise that rejects in <ms> milliseconds
        return new Promise((resolve) => {
            setTimeout(() => {
                // eslint-disable-next-line @typescript-eslint/ban-ts-comment
                //@ts-ignore
                resolve();
            }, ms);
        });
    }
    static poll(logger, promiseFunction, totalPollingTime, pollIntervalMs) {
        const startTime = new Date().getMilliseconds();
        return promiseFunction().then(async (result) => {
            if (result) {
                return true;
            }
            else {
                if (BootstrapUtils.stopProcess) {
                    return Promise.resolve(false);
                }
                const endTime = new Date().getMilliseconds();
                const newPollingTime = Math.max(totalPollingTime - pollIntervalMs - (endTime - startTime), 0);
                if (newPollingTime) {
                    logger.info(`Retrying in ${pollIntervalMs / 1000} seconds. Polling will stop in ${newPollingTime / 1000} seconds`);
                    await BootstrapUtils.sleep(pollIntervalMs);
                    return this.poll(logger, promiseFunction, newPollingTime, pollIntervalMs);
                }
                else {
                    return false;
                }
            }
        });
    }
    static async generateConfiguration(
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    templateContext, copyFrom, copyTo, excludeFiles = [], includeFiles = []) {
        // Loop through all the files in the config folder
        await fs_1.promises.mkdir(copyTo, { recursive: true });
        const files = await fs_1.promises.readdir(copyFrom);
        await Promise.all(files.map(async (file) => {
            const fromPath = path_1.join(copyFrom, file);
            const toPath = path_1.join(copyTo, file);
            // Stat the file to see if we have a file or dir
            const stat = await fs_1.promises.stat(fromPath);
            if (stat.isFile()) {
                const isMustache = file.indexOf('.mustache') > -1;
                const destinationFile = toPath.replace('.mustache', '');
                const fileName = path_1.basename(destinationFile);
                const notBlacklisted = excludeFiles.indexOf(fileName) === -1;
                const inWhitelistIfAny = includeFiles.length === 0 || includeFiles.indexOf(fileName) > -1;
                if (notBlacklisted && inWhitelistIfAny) {
                    if (isMustache) {
                        const template = await BootstrapUtils.readTextFile(fromPath);
                        const renderedTemplate = this.runTemplate(template, templateContext);
                        await fs_1.promises.writeFile(destinationFile, destinationFile.toLowerCase().endsWith('.json')
                            ? BootstrapUtils.formatJson(renderedTemplate)
                            : renderedTemplate);
                    }
                    else {
                        await fs_1.promises.copyFile(fromPath, destinationFile);
                    }
                    await fs_1.promises.chmod(destinationFile, 0o600);
                }
            }
            else if (stat.isDirectory()) {
                await fs_1.promises.mkdir(toPath, { recursive: true });
                await this.generateConfiguration(templateContext, fromPath, toPath, excludeFiles, includeFiles);
            }
        }));
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    static runTemplate(template, templateContext) {
        try {
            const compiledTemplate = Handlebars.compile(template);
            return compiledTemplate(templateContext);
        }
        catch (e) {
            const securedTemplate = Utils_1.Utils.secureString(template);
            const securedContext = Utils_1.Utils.secureString(BootstrapUtils.toYaml(templateContext));
            const securedMessage = Utils_1.Utils.secureString(e.message || 'Unknown');
            const message = `Unknown error rendering template. Error: ${securedMessage}\nTemplate:\n${securedTemplate}.`;
            throw new Error(`${message}\nContext: \n${securedContext}`);
        }
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    static async writeYaml(path, object, password) {
        const yamlString = this.toYaml(password ? CryptoUtils_1.CryptoUtils.encrypt(object, BootstrapUtils.validatePassword(password)) : object);
        await BootstrapUtils.writeTextFile(path, yamlString);
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    static pruneEmpty(obj) {
        return (function prune(current) {
            _.forOwn(current, (value, key) => {
                if (_.isUndefined(value) || _.isNull(value) || _.isNaN(value) || (_.isObject(value) && _.isEmpty(prune(value)))) {
                    delete current[key];
                }
            });
            // remove any leftover undefined values from the delete
            // operation on an array
            if (_.isArray(current))
                _.pull(current, undefined);
            return current;
        })(_.cloneDeep(obj)); // Do not modify the original object, create a clone instead
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    static toYaml(object) {
        return yaml.safeDump(object, { skipInvalid: true, indent: 4, lineWidth: 140, noRefs: true });
    }
    static fromYaml(yamlString) {
        return yaml.safeLoad(yamlString);
    }
    static loadYaml(fileLocation, password) {
        const object = this.fromYaml(this.loadFileAsText(fileLocation));
        if (password) {
            BootstrapUtils.validatePassword(password);
            try {
                return CryptoUtils_1.CryptoUtils.decrypt(object, password);
            }
            catch (e) {
                throw new KnownError(`Cannot decrypt file ${fileLocation}. Have you used the right password?`);
            }
        }
        else {
            if (password !== false && CryptoUtils_1.CryptoUtils.encryptedCount(object) > 0) {
                throw new KnownError(`File ${fileLocation} seems to be encrypted but no password has been provided. Have you entered the right password?`);
            }
        }
        return object;
    }
    static loadFileAsText(fileLocation) {
        return fs_1.readFileSync(fileLocation, 'utf8');
    }
    static async writeTextFile(path, text) {
        const mkdirParentFolder = async (fileName) => {
            const parentFolder = path_1.dirname(fileName);
            if (parentFolder) {
                return fs_1.promises.mkdir(parentFolder, { recursive: true });
            }
        };
        await mkdirParentFolder(path);
        await fs_1.promises.writeFile(path, text, 'utf8');
    }
    static async readTextFile(path) {
        return fs_1.promises.readFile(path, 'utf8');
    }
    static add(a, b) {
        if (_.isNumber(a) && _.isNumber(b)) {
            return Number(a) + Number(b);
        }
        if (typeof a === 'string' && typeof b === 'string') {
            return a + b;
        }
        return '';
    }
    static minus(a, b) {
        if (!_.isNumber(a)) {
            throw new TypeError('expected the first argument to be a number');
        }
        if (!_.isNumber(b)) {
            throw new TypeError('expected the second argument to be a number');
        }
        return Number(a) - Number(b);
    }
    static computerMemory(percentage) {
        return (os_1.totalmem() * percentage) / 100;
    }
    static toAmount(renderedText) {
        const numberAsString = (renderedText + '').split("'").join('');
        if (!numberAsString.match(/^\d+$/)) {
            throw new Error(`'${renderedText}' is not a valid integer`);
        }
        return (numberAsString.match(/\d{1,3}(?=(\d{3})*$)/g) || [numberAsString]).join("'");
    }
    static toHex(renderedText) {
        if (!renderedText) {
            return '';
        }
        const numberAsString = BootstrapUtils.toSimpleHex(renderedText);
        return '0x' + (numberAsString.match(/\w{1,4}(?=(\w{4})*$)/g) || [numberAsString]).join("'");
    }
    static toSimpleHex(renderedText) {
        if (!renderedText) {
            return '';
        }
        return renderedText.toString().split("'").join('').replace(/^(0x)/, '');
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    static toJson(object) {
        return JSON.stringify(object, null, 2);
    }
    static formatJson(string) {
        // Validates and format the json string.
        try {
            return JSON.stringify(JSON.parse(string), null, 2);
        }
        catch (e) {
            throw new Error(`${e.message}:JSON\n ${string}`);
        }
    }
    static splitCsv(object) {
        return (object || '')
            .split(',')
            .map((string) => string.trim())
            .filter((string) => string);
    }
    static toSeconds(serverDuration) {
        return symbol_sdk_1.DtoMapping.parseServerDuration(serverDuration).seconds();
    }
    // eslint-disable-next-line @typescript-eslint/explicit-module-boundary-types
    static migrate(logger, entityName, versioned, migrations = []) {
        if (!versioned) {
            return versioned;
        }
        const currentVersion = migrations.length + 1;
        versioned.version = versioned.version || 1;
        if (versioned.version == currentVersion) {
            return versioned;
        }
        logger.info(`Migrating object ${entityName} from version ${versioned.version} to version ${currentVersion}`);
        if (versioned.version > currentVersion) {
            throw new Error(`Current data version is ${versioned.version} but higher version is ${currentVersion}`);
        }
        const migratedVersioned = migrations.slice(versioned.version - 1).reduce((toMigrateData, migration) => {
            if (toMigrateData === undefined) {
                logger.info(`data to migrate is undefined, ignoring migration ${migration.description}`);
                return undefined;
            }
            logger.info(`Applying migration ${migration.description}`);
            return migration.migrate(toMigrateData);
        }, versioned);
        migratedVersioned.version = currentVersion;
        logger.info(`Object ${entityName} migrated to version ${currentVersion}`);
        return migratedVersioned;
    }
    static getNetworkIdentifier(networkType) {
        return BootstrapUtils.getNetworkName(networkType);
    }
    static getNetworkName(networkType) {
        switch (networkType) {
            case symbol_sdk_1.NetworkType.MAIN_NET:
                return 'mainnet';
            case symbol_sdk_1.NetworkType.TEST_NET:
                return 'testnet';
        }
        throw new Error(`Invalid Network Type ${networkType}`);
    }
    static resolveWorkingDirPath(workingDir, path) {
        if (path_1.isAbsolute(path)) {
            return path;
        }
        else {
            return path_1.join(workingDir, path);
        }
    }
    static createDerFile(privateKey, file) {
        fs_1.writeFileSync(file, symbol_sdk_1.Convert.hexToUint8(BootstrapUtils.toAns1(privateKey)));
    }
    static isYmlFile(string) {
        return string.toLowerCase().endsWith('.yml') || string.toLowerCase().endsWith('.yaml');
    }
    static validatePassword(password) {
        const passwordMinSize = 4;
        if (password.length < passwordMinSize) {
            throw new KnownError(`Password is too short. It should have at least ${passwordMinSize} characters!`);
        }
        return password;
    }
}
exports.BootstrapUtils = BootstrapUtils;
BootstrapUtils.stopProcess = false;
BootstrapUtils.onProcessListener = (() => {
    process.on('SIGINT', () => {
        BootstrapUtils.stopProcess = true;
    });
})();
//HANDLEBARS READY FUNCTIONS:
BootstrapUtils.initialize = (() => {
    Handlebars.registerHelper('toAmount', BootstrapUtils.toAmount);
    Handlebars.registerHelper('toHex', BootstrapUtils.toHex);
    Handlebars.registerHelper('toSimpleHex', BootstrapUtils.toSimpleHex);
    Handlebars.registerHelper('toSeconds', BootstrapUtils.toSeconds);
    Handlebars.registerHelper('toJson', BootstrapUtils.toJson);
    Handlebars.registerHelper('splitCsv', BootstrapUtils.splitCsv);
    Handlebars.registerHelper('add', BootstrapUtils.add);
    Handlebars.registerHelper('minus', BootstrapUtils.minus);
    Handlebars.registerHelper('computerMemory', BootstrapUtils.computerMemory);
})();
